unit RunScri;
{===============================================================================
 "Paradox database version upgrader" demo application.

         Roman Krejci-RKS,
         Pujmanove 1552,
  140 00 Prague 4,
         Czech Republic

 http://www.rksolution.cz,
 EMail INFO@rksolution.cz
 -------------------------------------------------------------------------------

 This unit implements routines for inplace Paradox database version upgrade.

 The core routine is

 Function RunPXUPScriptLines(const DLLDir,
                                  Dbsename,
                                  LogFilePath: string;
                                  szDriver   : pChar;
                                  KnownPasswords,
                                  ScriptLines      : tStrings;
                                  clbck            : tPXUPGRCallbackFunc
                                  ): Integer;

 that applies script containing PXUPGR commands to database "Dbsename".


 Arguments
    DLLDir         : string;   ... Folder from where pxupgr.dll should be loaded.
                                   If you specify empty string, default system
                                   search path is used.
    Dbsename       : string;   ... Database name for the database the script should be
                                   applied to. You can specify either an Alias,
                                   or physical database path
    LogFilePath    : string;   ... Complete path to Log file.
                                   You can specify empty string if you do not want
                                   to use LOG file.
    szDriver       : pChar;    ... szDriverType argument tells what type of database the script is
                                   applied on:
                                   'PARADOX'  (szPARADOX constant in BDE unit) for Paradox database,
                                   'DBASE'    (szDBASE constant in BDE unit) for DBASE database,
                                   'ASCIIDRV' (szASCII constant in BDE unit) for ASCII database.
                                   Current version supports only PARADOX driver.
    KnownPasswords : tStrings; ... String list containing known passwords for the upgraded
                                   database. Can be NIL.
    Script lines   : tStrings; ... String list containing script lines with PXUPGR commands,
                                   such as

                                   'tbl_create("BIOLIFE","ascii",4,"Species No",FLDFLOAT,FLDSTDEFAULT,1,0)'.

                                   Only one command is allowed per line. Script can contain
                                   "remark" lines - these lines should start with ':' (colon)
                                   character.

    clbck          : tPXUPGRCallbackFunc; ... Callback function for feedback (gauge?) purposes.

 Returns:

    If an error occurs during execution, routine terminates and returns
    corresponding error code. Other than that, it can return either pxuErr_None or
    pxuErr_AlreadyUpdated, depending on whether changes specified by the script
    are already present in the database or not.

 Log file format

    Routine writes into log file by using following format:
    BEFORE executing a command, line is written that contains '#' character followed
    by the line number of the script line containing the command (remark and empty
    lines do not count). The rest of this line, separated by at least one space character,
    can hold additional user specified information (only line number is important for
    correct routine behaviour). By default, the rest consists of Date/Time of the execution
    plus text of the script line.
    After that, actual script line containing PXUPGR command is executed. AFTER executing
    the command, line is written into LOG file that either states success, or describes
    error if any error ocurred.

   EXAMPLE:

    #1 - 23.9.1999 15:04:25: tbl_create("BIOLIFE","ascii",4,"Species No",FLDFLOAT,FLDSTDEFAULT,1,0)
    Success

      OR

    #1 - 23.9.1999 15:04:25: tbl_create("BIOLIFE","ascii",4,"Species No",FLDFLOAT,FLDSTDEFAULT,1,0)
    Error in executing command: Out of memory


    The format of the log file is important for the logic the routine uses in order
    to determine what commands were already applied to the database in the past.
    Before trying to apply the script to the database, the routine iterates
    LOG file and searches for the highest number following leading # character
    ocurring on the line that is followed by the line containing leading word 'Success'.
    Script line with this number (remark and empty lines do not count) is considered
    the last successfuly applied command.
    The routine skips this command and all preceding commands in order to
    avoid repeated command execution - this speeds up the process significantly.
    It is therefore important to keep in mind that constants

    pxuMsg_Success = 'Success';
    pxuMsg_Err_AlreadyUpdated   = pxuMsg_Success +' - Database already updated';
    pxuMsg_Err_None             = pxuMsg_Success;

    in PxUpgrH unit are crucial:  both  pxuMsg_Err_AlreadyUpdated  and pxuMsg_Err_None
    strings must start with pxuMsg_Success string.

    If, for any reason, the log file is deleted, the routine tries to apply the whole
    script to the database. This, however, should have no other effect than
    unnecessary delay in application startup, because  pxupExecLine function exported
    by pxupgr.dll will recognize duplicated actions (in which case it will return
    pxuErr_AlreadyUpdated code).

  !!! IMPORTANT NOTE FOR DATABASE BACKUP STRATEGY !!!
    Because the log file does now constitute the integral part of database structure,
    do not forget to include it into database backups.
   ===============================================================================}

interface
uses
  Windows,BDE,Sysutils,Classes,DB,DBTables, Controls, Dialogs,
  PxUpgrH, //  PXUPGR.DLL import unit
  Pswd;    //  dialog for password retrieval

Const
{ Application defined errors }
  AppErr_UserAbort                 = pxuErr_User+1; { operation aborted by user }

 {$I RUNSCRI.INC} { contains string constants to be translated }


Type
  pPXUPCallbackRecord=^tPXUPCallbackRecord;
  tPXUPCallbackRecord = record
    Dbsename,             { Database name of the database the script is applied to }
    ExecMsg,              { Error message for ExecResult member }
    ScriptLine: string;   { Current script line }
    BeforeApply: boolean; { True: callback is called BEFORE passing script line to pxupExecLine.
                                  ExecResult member is zero
                            False: callback is called AFTER passing script line to pxupExecLine -
                                   ExecResult member then holds error code returned by pxupExecLine
                                   and ExecMsg holds coresponding Success/Error message }
    ExecResult: Integer;  { Result of pxupExecLine function - valid only if BeforeApply member is false }
    LinesTotal,           { Total count of NON-remark and NON-empty ("executable") lines in the script }
    LinesApplicable,      { Added in ver 1.02. LinesTotal MINUS number of "executable" lines marked 
                            in the log file as already succesfuly applied in the past }
    CurrentLine,          { Line number of current script line. Remark and empty lines do not count }
    LinesDone:  integer;  { Total count of so far executed script lines - this can differ from CurrentLine
                            if some script lines were skipped because they are marked in the log file as
                            already succesfuly applied in the past }
  end;

{ Callback function prototype. Should return false if script execution should be interrupted }
  tPXUPGRCallbackFunc = function(pRec: pPXUPCallbackRecord): boolean;


Function RunPXUPScriptLines(const DLLDir,                { from where PXUPGR.DLL should be loaded }
                                  Dbsename,              { Database the script should be applied to.
                                                           Can be either Alias or physical path }
                                  LogFilePath: string;   { Complete path to log file. can be empty string
                                                           if no LOG file should be used }
                                  szDriver   : pChar;    { driver type }
                            KnownPasswords,              { String list holding known passwords for the database }
                            ScriptLines      : tStrings; { String list containing script lines }
                            clbck            : tPXUPGRCallbackFunc { Callback function. Can be nil }
                            ): Integer;


{ Following function differs from RunPXUPScriptLines in that, instead of
  ScriptLines: tStrings; argument, it accepts path to the script file }
Function RunPXUPScript(const DLLDir,
                             Dbsename,
                             LogFilePath: string;
                             szDriver   : pChar;
                             KnownPasswords: tStrings;
                             ScriptFilePath: string;    // path to script file
                             clbck : tPXUPGRCallbackFunc): Integer;


implementation

Const
  RemChar = ':';
  LogCommandChar = '#';

{ Exclusively opens text file. File must exist. If exclusive open fails,
  message is popped up that asks the user whether to retry or abort.
  In case user decides to abort, EAbort is raised. }
Procedure TextOpenExclusive(var t: textFile;
                           const sFname: string; { File must exist }
                           forRead: Boolean {True=Reset,False=Append});
var h: integer; // file handle
    StayInLoop: Boolean; // loop termination flag
Begin
  assignFile(t, sFname);
  Repeat
    StayInLoop := false;
    h := fileOpen(sFname, fmShareExclusive);
    if h > 0 then begin
      fileClose(h);
      if ForRead then Reset(t) else Append(t);
    end else
      if MessageDlg(sScriptProcessedByOtherStation,
        mtConfirmation,[mbYes,mbCancel], 0) <> mrYes
        then Sysutils.Abort
           else StayInLoop := True;
  Until not StayInLoop;
End;

{ Following function retrieves line number of the last successfuly executed script line
  from the log file (empty lines and lines starting with RemChar do not count).
  The routine iterates LOG file and searches for the highest number following leading # character
  ocurring on the line that is followed by the line containing leading word 'Success'.
  Line numbers are 1 based. Zero result means that no script line was logged with
  'success' outcome so far}
Function GetLastSuccessLineNumber(const LogFilePath: string): Integer;
var t: Text;       // log file
    Line: string;  // line read from log file
    pos32,         // position of 1st ' ' character in Line
    Current,       // current line number of the logged script line
    Code: Integer; // param for VAL routine
begin
  result := 0;
  if (LogFilePath <> '') and   // Log file specified ...
     FileExists(LogFilePath)   // ... and already created
  then begin
    TextOpenExclusive(t,LogFilePath,True{file will be RESET});
    Try
      Current := 0;
      while not EOF(t) do begin
        Readln(t,Line);
        if Current = 0 then begin
        { Current = 0 => previous line DOES NOT contain logged script command.
          Perhaps the actual line does... }
          if (length(Line) > 0) and (Line[1]=LogCommandChar) then begin
         { extract line number from the Log file line: }
            pos32 := pos(' ',Line);
            if pos32 > 0 then begin
              Line := copy(Line,2,pos32-2);
              VAL(Line,Current,code);
              if Code <> 0 then Current := 0; // superfluous?
            end;
          end;
        end else begin // Current <> 0 => previous line contains logged script command
          if Pos(pxuMsg_Success,Line) = 1 then begin // Log file states success...
            if Current > Result then Result := Current;
          end;
          Current := 0;
        end;
      end;
    Finally
      closefile(t);
    End
  end;
end;

Function RunPXUPScriptLines(const DLLDir,
                             Dbsename,
                             LogFilePath: string;
                             szDriver   : pChar;
                             KnownPasswords,
                             ScriptLines: tStrings;
                       clbck: tPXUPGRCallbackFunc): Integer;
var
  t: TextFile;    // Log file
  tOpen: Boolean; // Log file opened flag

    procedure InitLogFile;
    begin
      if not tOpen and (LogFilePath <> '') then begin
        if not FileExists(LogFilePath) then
        with tFileStream.Create(LogFilePath,fmOpenWrite or fmShareDenyNone or fmCreate)
          do free;
        TextOpenExclusive(t,LogFilePath,False{file will be APPENDed});
        tOpen := true;
      end;
    end;

    procedure DoneLogFile;
    begin
      if tOpen then CloseFile(t);
      tOpen := false;
    end;

    procedure WriteLnToLogFile(const msg: string);
    begin
      if tOpen then Writeln(t,msg);
    end;

var LastSuccessLine,               // retrieved from LOG file - last sucessfuly applied script line
    i: Integer;                    // counter over ScriptLines.Strings
    clbckRec: tPXUPCallbackRecord; // info for callback purposes
    OldPswdCallback: Pointer;      // previously registered callback func for unknown password retrieval
    LogLine: string;               // Line written into LOG file
begin
  tOpen := false; // Log file not opened so far
  Result := pxuErr_None; // assume success
  if ScriptLines <> nil then Try
    FillChar(clbckRec,SizeOf(clbckRec),#0);
 { Get total number of "executable" script lines }
    For i := 0 to ScriptLines.Count-1 do
      if length(Sysutils.Trim(ScriptLines[i])) > 0 then
        if ScriptLines[i][1] <> RemChar then
          inc(clbckRec.LinesTotal);
 { Apply "executable" script lines to database }
    if clbckRec.LinesTotal > 0 then begin
 { following call has blocking effect if LogFilepath <> '' and the database is
   currently being upgraded by another station: }
      LastSuccessLine := GetLastSuccessLineNumber(LogFilePath);
      clbckRec.LinesApplicable := clbckRec.LinesTotal - LastSuccessLine;
      if clbckRec.LinesApplicable > 0 then begin
        Result := pxupgrInit(DllDir);
        if Result = pxuErr_None then Try
 { Register callback for unknown password retrieval:}
          OldPswdCallback := pxupRegisterPswdCallback(@UnknownPasswordCallback); // from unit Pswd
          Try
 { Fetch known database passwords into PXUPGR.DLL }
            if KnownPasswords <> nil then
              For i := 0 to KnownPasswords.Count-1 do pxupAddPassword(pChar(KnownPasswords[i]));
 { following call has blocking effect if LogFilepath <> '' and the database is
   currently being upgraded by another station: }
            InitLogFile;
            clbckRec.Dbsename := Dbsename;
            For i := 0 to ScriptLines.Count-1 do begin
              if length(Sysutils.Trim(ScriptLines[i])) > 0 then begin  // line not empty
                if ScriptLines[i][1] <> RemChar then begin             // line is not remark line
                  inc(clbckRec.CurrentLine);
                  if clbckRec.CurrentLine > LastSuccessLine then begin
                    clbckRec.BeforeApply := True;
                    clbckRec.ExecResult := 0;
                    clbckRec.ExecMsg := '';
                    clbckRec.ScriptLine := Sysutils.Trim(ScriptLines[i]);
 { Here the callback routine is called for the first time.
   clbckRec.LinesTotal and clbckRec.LinesApplicable are guaranteed to be > 0}
                    if assigned(clbck) and not clbck(@clbckRec) then Sysutils.Abort;
                    LogLine := Format(LogCommandChar+'%d ',[clbckRec.CurrentLine]); // mandatory
                    LogLine := LogLine + '- '+dateTimeToStr(Now);
 { comment out following line if you don't want ScriptLine text occurring in the LOG }
                    LogLine := LogLine + ': '+clbckRec.ScriptLine;
                    WriteLnToLogFile(LogLine);
                    Result := pxupExecLine(pChar(Dbsename),szDriver,pChar(clbckRec.ScriptLine));
                    clbckRec.ExecResult := Result;
                    inc(clbckRec.LinesDone);
                    clbckRec.BeforeApply := False;
                    clbckRec.ExecMsg := GetPXUPErrorMsg(clbckRec.ExecResult);
                    LogLine := clbckRec.ExecMsg;
                    WriteLnToLogFile(LogLine);
                    if assigned(clbck) and not clbck(@clbckRec) then
                      if Result <= 0 then Sysutils.Abort;  // abort only if command executed OK
                    if Result > 0 then break; // error - break loop and report error code
                  end   { if  }
                end; { if  }
              end; { if  }
            end; { For }
          Finally
            pxupRegisterPswdCallback(OldPswdCallback); // reset old password callback
          End;
        Finally
          DoneLogFile;
          pxupgrDone;
        end
      end else
       if clbckRec.LinesApplicable < 0 then
 { This occurs if the LOG file indicates higher number of succesfuly executed commands
   than present in ScriptLines list. Most often this happens if the end-user downgraded
   to older application version with shorter upgrade script. }
          showmessage(sWarningVersionMismatch);
    End;
  Except
    On EAbort do Result := AppErr_UserAbort;
  End
end;



Function RunPXUPScript(const DLLDir,
                             Dbsename,
                             LogFilePath: string;
                             szDriver   : pChar;
                             KnownPasswords: tStrings;
                             ScriptFilePath: string;
                             clbck : tPXUPGRCallbackFunc): Integer;
Var
  ScriptLines: tStringList;
begin
  ScriptLines := tStringList.Create;
  Try
    if (ScriptFilePath <> '') and FileExists(ScriptFilePath) then
      ScriptLines.LoadFromFile(ScriptFilePath);
    Result := RunPXUPScriptLines(DLLDir,
                                 Dbsename,
                                 LogFilePath,
                                 szDriver,
                                 KnownPasswords,
                                 ScriptLines,
                                 clbck);
  Finally
    ScriptLines.Free;
  End;
end;


end.
